"use strict";
var UnRecursiveFileSystemWatcher_1;
Object.defineProperty(exports, "__esModule", { value: true });
exports.UnRecursiveFileSystemWatcher = void 0;
const tslib_1 = require("tslib");
const fs_extra_1 = tslib_1.__importStar(require("fs-extra"));
const debounce_1 = tslib_1.__importDefault(require("lodash/debounce"));
const di_1 = require("@opensumi/di");
const ide_core_node_1 = require("@opensumi/ide-core-node");
const index_1 = require("../../common/index");
const file_change_collection_1 = require("../file-change-collection");
const { join, basename, normalize } = ide_core_node_1.path;
let UnRecursiveFileSystemWatcher = UnRecursiveFileSystemWatcher_1 = class UnRecursiveFileSystemWatcher {
    constructor(excludes = []) {
        this.excludes = excludes;
        this.WATCHER_HANDLERS = new Map();
        // 收集发生改变的文件
        this.changes = new file_change_collection_1.FileChangeCollection();
        this.toDispose = new ide_core_node_1.DisposableCollection(ide_core_node_1.Disposable.create(() => this.setClient(undefined)));
        /**
         * Fires file changes to clients.
         * It is debounced in the case if the filesystem is spamming to avoid overwhelming clients with events.
         */
        this.fireDidFilesChanged = (0, debounce_1.default)(() => this.doFireDidFilesChanged(), 100);
        this.logger = this.loggerManager.getLogger(ide_core_node_1.SupportLogNamespace.Node);
    }
    dispose() {
        this.toDispose.dispose();
        this.WATCHER_HANDLERS.clear();
    }
    /**
     * 查找某个路径是否已被监听
     * @param watcherPath
     */
    checkIsAlreadyWatched(watcherPath) {
        for (const [watcherId, watcher] of this.WATCHER_HANDLERS) {
            if (watcherPath.indexOf(watcher.path) === 0) {
                return watcherId;
            }
        }
    }
    async doWatch(basePath) {
        try {
            const watcher = (0, fs_extra_1.watch)(basePath);
            this.logger.log('start watching', basePath);
            const isDirectory = fs_extra_1.default.lstatSync(basePath).isDirectory();
            const docChildren = new Set();
            let signalDoc = '';
            if (isDirectory) {
                try {
                    for (const child of fs_extra_1.default.readdirSync(basePath)) {
                        const base = join(basePath, String(child));
                        if (!fs_extra_1.default.lstatSync(base).isDirectory()) {
                            docChildren.add(child);
                        }
                    }
                }
                catch (error) {
                    this.logger.error(error);
                }
            }
            else {
                signalDoc = basename(basePath);
            }
            // 开始走监听流程
            watcher.on('error', (code, signal) => {
                this.logger.error(`Failed to watch ${basePath} for changes using fs.watch() (${code}, ${signal})`);
                watcher.close();
            });
            watcher.on('change', (type, filename) => {
                if (this.isTemporaryFile(filename)) {
                    return;
                }
                // 对传入的raw做一个统一处理
                let changeFileName = '';
                if (filename) {
                    changeFileName = filename;
                    if (ide_core_node_1.isMacintosh) {
                        changeFileName = normalize(changeFileName);
                    }
                }
                if (!filename || (type !== 'change' && type !== 'rename')) {
                    return;
                }
                const changePath = join(basePath, changeFileName);
                if (isDirectory) {
                    setTimeout(async () => {
                        // 监听的目录如果是文件夹，那么只对其下面的文件改动做出响应
                        if (docChildren.has(changeFileName)) {
                            if ((type === 'rename' || type === 'change') && changeFileName === filename) {
                                const fileExists = fs_extra_1.default.existsSync(changePath);
                                if (fileExists) {
                                    this.pushUpdated(changePath);
                                }
                                else {
                                    docChildren.delete(changeFileName);
                                    this.pushDeleted(changePath);
                                }
                            }
                        }
                        else if (fs_extra_1.default.pathExistsSync(changePath)) {
                            if (!fs_extra_1.default.lstatSync(changePath).isDirectory()) {
                                this.pushAdded(changePath);
                                docChildren.add(changeFileName);
                            }
                        }
                    }, UnRecursiveFileSystemWatcher_1.FILE_DELETE_HANDLER_DELAY);
                }
                else {
                    setTimeout(async () => {
                        if (changeFileName === signalDoc) {
                            if (fs_extra_1.default.pathExistsSync(basePath)) {
                                this.pushUpdated(basePath);
                            }
                            else {
                                this.pushDeleted(basePath);
                                signalDoc = '';
                            }
                        }
                    }, UnRecursiveFileSystemWatcher_1.FILE_DELETE_HANDLER_DELAY);
                }
            });
        }
        catch (error) {
            this.logger.error(`Failed to watch ${basePath} for change using fs.watch() (${error.toString()})`);
        }
    }
    async watchFileChanges(uri) {
        const basePath = ide_core_node_1.FileUri.fsPath(uri);
        const exist = await fs_extra_1.default.pathExists(basePath);
        let watcherId = this.checkIsAlreadyWatched(basePath);
        if (watcherId) {
            return watcherId;
        }
        watcherId = UnRecursiveFileSystemWatcher_1.WATCHER_SEQUENCE++;
        const disposables = new ide_core_node_1.DisposableCollection(); // 管理可释放的资源
        let watchPath = '';
        if (exist) {
            const stat = await fs_extra_1.default.lstatSync(basePath);
            if (stat) {
                watchPath = basePath;
            }
        }
        else {
            this.logger.warn('This path does not exist. Please try again');
        }
        disposables.push(await this.start(watchPath));
        this.toDispose.push(disposables);
        return watcherId;
    }
    async start(basePath) {
        const disposables = new ide_core_node_1.DisposableCollection();
        if (!(await fs_extra_1.default.pathExists(basePath))) {
            return disposables;
        }
        const realPath = await fs_extra_1.default.realpath(basePath);
        const tryWatchDir = async (retryDelay = 1000) => {
            try {
                this.doWatch(realPath);
            }
            catch (error) {
                await new Promise((resolve) => {
                    setTimeout(resolve, retryDelay);
                });
            }
            return undefined;
        };
        await tryWatchDir();
        return disposables;
    }
    unwatchFileChanges(watcherId) {
        const watcher = this.WATCHER_HANDLERS.get(watcherId);
        if (watcher) {
            this.WATCHER_HANDLERS.delete(watcherId);
            watcher.disposable.dispose();
        }
        return Promise.resolve();
    }
    pushAdded(path) {
        this.pushFileChange(path, index_1.FileChangeType.ADDED);
    }
    pushUpdated(path) {
        this.pushFileChange(path, index_1.FileChangeType.UPDATED);
    }
    pushDeleted(path) {
        this.pushFileChange(path, index_1.FileChangeType.DELETED);
    }
    pushFileChange(path, type) {
        const uri = ide_core_node_1.FileUri.create(path).toString();
        this.changes.push({ uri, type });
        this.fireDidFilesChanged();
    }
    doFireDidFilesChanged() {
        const changes = this.changes.values();
        this.changes = new file_change_collection_1.FileChangeCollection();
        const event = { changes };
        if (this.client) {
            this.client.onDidFilesChanged(event);
        }
    }
    setClient(client) {
        if (client && this.toDispose.disposed) {
            return;
        }
        this.client = client;
    }
    isTemporaryFile(path) {
        if (path) {
            if (/\.\d{7}\d+$/.test(path)) {
                // write-file-atomic 源文件xxx.xx 对应的临时文件为 xxx.xx.22243434
                // 这类文件的更新应当完全隐藏掉
                return true;
            }
        }
        return false;
    }
};
UnRecursiveFileSystemWatcher.WATCHER_SEQUENCE = 1;
UnRecursiveFileSystemWatcher.FILE_DELETE_HANDLER_DELAY = 500;
tslib_1.__decorate([
    (0, di_1.Autowired)(ide_core_node_1.ILogServiceManager)
    // 一个 symbol 关键字，内容是 ILogServiceManager
    ,
    tslib_1.__metadata("design:type", Object)
], UnRecursiveFileSystemWatcher.prototype, "loggerManager", void 0);
UnRecursiveFileSystemWatcher = UnRecursiveFileSystemWatcher_1 = tslib_1.__decorate([
    (0, di_1.Injectable)({ multiple: true }),
    tslib_1.__param(0, (0, di_1.Optional)()),
    tslib_1.__metadata("design:paramtypes", [Array])
], UnRecursiveFileSystemWatcher);
exports.UnRecursiveFileSystemWatcher = UnRecursiveFileSystemWatcher;
//# sourceMappingURL=file-service-watcher.js.map